@@ -1,4 +1,5 @@
 import asyncio
+import functools
 import random
 import time
 from concurrent.futures import ThreadPoolExecutor
@@ -57,7 +58,8 @@ class ProxyManager:
                  num_executors=32,
                  executor_queue_size=1000,
                  verbose=False,
-                 new_session_fn=lambda: requests.Session()):
+                 new_session_fn=lambda: requests.Session(),
+                 loop=asyncio.get_event_loop()):
         """
         Create a new proxy manager. The manager has the internal thread workers, so to make it properly
         work, use it as a context manager. ex) with ProxyManager(...) as proxy_manager: ...
@@ -74,12 +76,13 @@ class ProxyManager:
         self._proxy_info_dict = proxy_info_dict
         self._verbose = verbose
         self._new_session_fn = new_session_fn
+        self._loop = loop
 
         self._pool = self._init_pool(proxy_info_dict)
 
-        self._executor_running = False
+        # self._executor_running = False
         self._executor_workers = num_executors
-        self._executor_queue = Queue(executor_queue_size)
+        # self._executor_queue = Queue(executor_queue_size)
         self._executor = ThreadPoolExecutor(max_workers=self._executor_workers)
 
         self._metrics_lock = Lock()
@@ -114,14 +117,14 @@ class ProxyManager:
                                  success_criterion: Callable[[Response], bool] = lambda response: True) -> Session:
         return SessionProxy(self, success_criterion)
 
-    def proxy_session(self,
-                      fn: Callable[[Session, tuple], Optional[Any]],
-                      callback_async: Callable[[
-                          Optional[Any]], None] = lambda x: None,
-                      exception_handle: Callable[[
-                          ProxyInfo, Exception], None] = lambda info, ex: print(info, ex),
-                      args: tuple = (),
-                      wait=False):
+    async def proxy_session(self,
+                            fn: Callable[[Session, tuple], Optional[Any]],
+                            callback_async: Callable[[
+                                Optional[Any]], None] = lambda x: None,
+                            exception_handle: Callable[[
+                                ProxyInfo, Exception], None] = lambda info, ex: print(info, ex),
+                            args: tuple = (),
+                            wait=False):
         """
         Work on the given function using the randomly chosen Session from the proxy pool. 'fn' will be retried
         indefinitely until success in order to try all available proxies.
@@ -135,19 +138,32 @@ class ProxyManager:
         :return: if 'wait' is True, the return value of 'fn' will be returned; None otherwise
         """
 
-        def task_fn():
+        end_event = asyncio.Event()
+
+        async def task_fn():
             success = False
 
-            while not success:
-                with self._pool.poll() as proxy:
+            while not end_event.set() and not success:
+                with await self._pool.poll() as proxy:
                     session = self._new_session_fn()
 
                     session.proxies.update(proxy.info().as_requests_dict())
 
                     timer = SimpleTimer()
                     try:
-                        with timer:
-                            result = fn(session, *args)
+                        timer.start()
+                        result_future = self._loop.run_in_executor(
+                            self._executor, functools.partial(fn, session, *args))
+
+                        while not result_future.done():
+                            await asyncio.sleep(0.1)
+                            if end_event.is_set():
+                                break
+
+                        if result_future.cancelled():
+                            break
+
+                        result = await result_future
 
                         if type(result) is tuple:
                             callback_async(*result)
@@ -170,42 +186,51 @@ class ProxyManager:
                         if self._verbose:
                             print("success rate:", self.success_rate())
 
-        task = Task(task_fn)
-        self._executor_queue.put(task, True)
+        # task = Task(task_fn)
+        # self._executor_queue.put(task, True)
+
+        done, pending = await asyncio.wait([asyncio.create_task(task_fn()) for _ in range(8)],
+                                           return_when=asyncio.FIRST_COMPLETED)
+        end_event.set()
+        for future in done:
+            if not future.done():
+                future.cancel()
+        future = next(iter(done))
 
         if wait:
-            return task.get()
+            # return task.get()
+            return await future
 
     async def start(self):
-        self._pool.start()
+        await self._pool.start()
 
-        def fn(index: int):
-            while self._executor_running:
-                try:
-                    task = self._executor_queue.get(False)
-                    task()
+        # def fn(index: int):
+        #     while self._executor_running:
+        #         try:
+        #             task = self._executor_queue.get(False)
+        #             task()
 
-                    self._executor_queue.task_done()
-                    continue
-                except Empty:
-                    time.sleep(3)
-                    continue
+        #             self._executor_queue.task_done()
+        #             continue
+        #         except Empty:
+        #             time.sleep(3)
+        #             continue
 
-            if self._verbose:
-                print("Worker {} is stopped working".format(index))
+        #     if self._verbose:
+        #         print("Worker {} is stopped working".format(index))
 
-        self._executor_running = True
-        for i in range(self._executor_workers):
-            self._executor.submit(fn, i)
+        # self._executor_running = True
+        # for i in range(self._executor_workers):
+        #     self._executor.submit(fn, i)
 
     async def stop(self, forced_shutdown=False):
-        if not forced_shutdown:
-            self._executor_queue.join()
+        # if not forced_shutdown:
+        #     self._executor_queue.join()
 
-        self._executor_running = False
+        # self._executor_running = False
         self._executor.shutdown(True)
 
-        self._pool.end()
+        await self._pool.end()
 
     def __enter__(self):
         loop = asyncio.get_event_loop()
@@ -228,7 +253,6 @@ class ProxyManager:
 class SessionProxy(Session):
     def __init__(self, manager: ProxyManager, success_criterion):
         super().__init__()
-        assert manager._executor_running, "manager hasn't started."
 
         self._manager = manager
         self._success_criterion = success_criterion
@@ -253,4 +277,5 @@ class SessionProxy(Session):
 
             return response
 
-        return self._manager.proxy_session(fn, wait=True)
+        loop = asyncio.get_event_loop()
+        return loop.run_until_complete(self._manager.proxy_session(fn, wait=True))
